from typing import List, Iterator, Set, Iterable, Optional, Dict, Callable, \
                   Any
import logging
import attr

from .core import FileLocationRange
from .problem import Problem

logger = logging.getLogger(__name__)  # type: logging.Logger


class Snippet(object):
    """
    Represents a code snippet that may be inserted into a program under
    repair.
    """
    @staticmethod
    def from_dict(d: Dict[str, Any]) -> 'Snippet':
        snippet = Snippet(d['content'])
        if 'locations' in d:
            for loc_s in d['locations']:
                loc = FileLocationRange.from_string(loc_s)
                snippet.locations.add(loc)
        return snippet

    def __init__(self, content: str) -> None:
        self.__content = content
        self.locations = set()  # type: Set[FileLocationRange]

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Snippet) and self.content == other.content

    def __hash__(self) -> int:
        return hash(self.content)

    @property
    def content(self) -> str:
        return self.__content

    @property
    def occurrences(self) -> int:
        return len(self.locations)

    def to_dict(self) -> Dict[str, Any]:
        d = {}  # type: Dict[str, Any]
        d['content'] = self.__content
        if self.locations:
            d['locations'] = [str(l) for l in self.locations]
        return d


class SnippetDatabase(object):
    @staticmethod
    def from_dict(d: List[Dict[str, Any]]) -> 'SnippetDatabase':
        snippets = [Snippet.from_dict(s) for s in d]
        return SnippetDatabase(snippets)

    def __init__(self,
                 snippets: Optional[Iterable[Snippet]] = None
                 ) -> None:
        """
        Constructs an empty snippet database.
        """
        if snippets is None:
            snippets = []

        self.__snippets = {}  # type: Dict[str, Snippet]
        for snippet in snippets:
            self.__snippets[snippet.content] = snippet

        self.__snippets_by_file = {}  # type: Dict[str, Set[Snippet]]
        for snippet in snippets:
            for location in snippet.locations:
                fn = location.filename
                if fn not in self.__snippets_by_file:
                    self.__snippets_by_file[fn] = set()
                self.__snippets_by_file[fn].add(snippet)

    def __iter__(self) -> Iterator[Snippet]:
        """
        Returns an iterator over the snippets contained in this databse.
        """
        yield from self.__snippets.values()

    def __len__(self) -> int:
        """
        Returns the number of unique snippets contained within the database.
        """
        return len(self.__snippets)

    def in_file(self, fn: str) -> Iterator[Snippet]:
        """
        Returns an iterator over all of the snippets that were sourced from
        a given file.
        """
        if fn in self.__snippets_by_file:
            yield from self.__snippets_by_file[fn]

    def add(self,
            content: str,
            *,
            origin: Optional[FileLocationRange] = None
            ) -> None:
        """
        Adds a snippet to this database in-place.

        Parameters:
            content: the content of the snippet.
            origin: an optional parameter that may be used to specify the
                origin of the snippet.

        Returns:
            nothing.
        """
        if content in self.__snippets:
            snippet = self.__snippets[content]
        else:
            snippet = Snippet(content)
            self.__snippets[content] = snippet

        if origin is not None:
            snippet.locations.add(origin)
            if origin.filename not in self.__snippets_by_file:
                self.__snippets_by_file[origin.filename] = set()
            self.__snippets_by_file[origin.filename].add(snippet)

    def to_dict(self) -> List[Dict[str, Any]]:
        return [s.to_dict() for s in self]


class SnippetFinder(object):
    def __init__(self, database: SnippetDatabase) -> None:
        self.__database = database

    def __next__(self) -> Iterator[Snippet]:
        yield from self.__database



from typing import Iterable, Iterator, Optional, List
from mypy_extensions import NoReturn
from timeit import default_timer as timer
import logging
import datetime
import threading
import time
import signal

import bugzoo

from .core import FileLine
from .candidate import Candidate
from .problem import Problem
from .outcome import OutcomeManager
from .exceptions import BuildFailure

logger = logging.getLogger(__name__)  # type: logging.Logger

__all__ = ['Searcher']


class Shutdown(Exception):
    pass


class Searcher(object):
    def __init__(self,
                 bugzoo: bugzoo.BugZoo,
                 problem: Problem,
                 candidates: Iterable[Candidate],
                 *,
                 threads: int = 1,
                 time_limit: Optional[datetime.timedelta] = None,
                 candidate_limit: Optional[int] = None
                 ) -> None:
        """
        Constructs a new searcher for a given source of candidate patches.

        Parameters:
            bugzoo: a connection to the BugZoo server that should be used to
                evaluate candidate patches.
            problem: a description of the problem.
            candidates: a source of candidate patches.
            threads: the number of threads that should be made available to
                the search process.
            time_limit: an optional limit on the amount of time given to the
                searcher.
            candidate_limit: an optional limit on the number of candidate
                patches that may be generated.
        """
        logger.debug("constructed searcher")
        assert time_limit is None or time_limit > datetime.timedelta(), \
            "if specified, time limit should be greater than zero."

        self.__bugzoo = bugzoo
        self.__problem = problem
        self.__candidates = candidates
        self.__time_limit = time_limit
        self.__candidate_limit = candidate_limit
        self.__num_threads = threads
        self.__outcomes = OutcomeManager()


        # records the time at which the current iteration begun
        self.__time_iteration_begun = None  # type: ignore

        self.__lock_candidates = threading.Lock()  # type: threading.Lock
        self.__counter_candidates = 0
        self.__counter_tests = 0
        self.__exhausted_candidates = False
        self.__time_running = datetime.timedelta()
        self.__error_occurred = False
        self.__searching = False
        self.__found_patches = []  # type: List[Candidate]
        self.__history = []  # type: List[Candidate]
        logger.debug("constructed searcher")

    @property
    def history(self) -> List[Candidate]:
        """
        Returns an ordered list of all of the candidate patches that have been
        explored by this search process.
        """
        return self.__history.copy()

    @property
    def outcomes(self) -> OutcomeManager:
        """
        Provides a log of the outcomes of candidate patch build attempts and
        test executions.
        """
        return self.__outcomes

    @property
    def paused(self) -> bool:
        """
        Indicates whether this searcher is paused.
        """
        return (self.__found_patches != []) or self.exhausted

    @property
    def exhausted(self) -> bool:
        """
        Indicates whether or not the resources available to this searcher have
        been exhausted.
        """
        if self.__error_occurred:
            return True
        if self.__exhausted_candidates:
            return True
        if self.__time_limit is not None:
            if self.time_running > self.__time_limit:  # type: ignore
                return True
        if self.__candidate_limit is not None:
            if self.__counter_candidates > self.__candidate_limit:
                return True
        return False

    @property
    def num_test_evals(self) -> int:
        """
        The number of test case evaluations that have been performed during
        this search process.
        """
        return self.__counter_tests

    @property
    def num_candidate_evals(self) -> int:
        """
        The number of candidate patches that have been evaluated over the
        course of this search process.
        """
        return self.__counter_candidates

    @property
    def time_limit(self) -> Optional[datetime.timedelta]:
        """
        An optional limit on the length of time that may be spent searching
        for patches.
        """
        return self.__time_limit

    @property
    def time_running(self) -> datetime.timedelta:
        """
        The amount of time that has been spent searching for patches.
        """
        duration_delta = self.__time_running
        if self.__searching:
            time_now = timer()
            iteration_secs = time_now - self.__time_iteration_begun  # type: ignore
            iteration_delta = datetime.timedelta(seconds=iteration_secs)  # type: datetime.timedelta
            duration_delta = duration_delta + iteration_delta
        logger.debug("time running: %.2f minutes", duration_delta.seconds / 60)
        return duration_delta

    def stop(self) -> None:
        # FIXME just use an event to communicate when to stop
        self.__error_occurred = True

    def __iter__(self) -> Iterator[Candidate]:
        return self

    def __next__(self) -> Candidate:
        """
        Searches for the next acceptable patch.

        Returns:
            the next patch that passes all tests.

        Raises:
            StopIteration: if the search space or available resources have
                been exhausted.
        """
        # if we have patches in the buffer, return those.
        if self.__found_patches:
            return self.__found_patches.pop()

        threads = [] # type: List[threading.Thread]

        # setup signal handlers to ensure that threads are cleanly killed
        # causes a Shutdown exception to be thrown in the search loop below
        if threading.current_thread() is threading.main_thread():
            logger.debug("attaching signal handlers")
            original_handler_sigint = signal.getsignal(signal.SIGINT)
            original_handler_sigterm = signal.getsignal(signal.SIGTERM)
            signal.signal(signal.SIGINT, lambda signum, frame: self.stop())
            signal.signal(signal.SIGTERM, lambda signum, frame: self.stop())
            logger.debug("attached signal handlers")
        else:
            logger.debug("not attaching signal handlers -- not inside main thread.")  # noqa: pycodestyle

        self.__time_iteration_begun = timer()  # type: ignore

        # TODO there's a bit of a bug: any patches that were read from the
        #   generator by the worker and were still stored in its
        #   `candidate` variable will be discarded. fixable by adding a
        #   buffer to `_try_next`.
        try:
            def worker(searcher: 'Searcher') -> None:
                while True:
                    if not searcher._try_next():
                        break

            self.__searching = True
            for _ in range(self.__num_threads):
                t = threading.Thread(target=worker, args=(self,))
                threads.append(t)
                t.start()

            for t in threads:
                t.join()
        except Shutdown:
            logger.info("Reached time limit without finding repair. Terminating search.")
            self.__error_occurred = True
        except Exception as e:
            self.__error_occurred = True
        finally:
            for t in threads:
                t.join()
            self.__searching = False
            if threading.current_thread() is threading.main_thread():
                logger.debug("restoring original signal handlers")
                signal.signal(signal.SIGINT, original_handler_sigint)
                signal.signal(signal.SIGTERM, original_handler_sigterm)
                logger.debug("restored original signal handlers")

        duration_iteration = timer() - self.__time_iteration_begun  # type: ignore
        self.__time_running += datetime.timedelta(seconds=duration_iteration)

        if self.__found_patches:
            return self.__found_patches.pop()
        raise StopIteration

    def _try_next(self) -> bool:
        """
        Evaluates the next candidate patch.

        Returns:
            a boolean indicating whether the calling thread should continue to
            evaluate candidate patches.
        """
        if self.paused:
            return False

        self.__lock_candidates.acquire()
        try:
            candidate = next(self.__candidates) # type: ignore
            self.__history.append(candidate)
            self.__counter_candidates += 1
        except StopIteration:
            logger.info("All candidate patches have been exhausted.")
            self.__exhausted_candidates = True
            return False
        finally:
            self.__lock_candidates.release()

        bz = self.__bugzoo
        mgr_src = self.__problem.sources

        patch = candidate.to_diff(self.__problem)
        lines_changed = candidate.lines_changed(self.__problem)  # type: List[FileLine]
        line_coverage_by_test = self.__problem.coverage
        logger.info("evaluating candidate: %s\n%s\n", candidate, patch)
        logger.debug("building candidate: %s", candidate)
        container = None
        time_build_start = timer()
        try:
            container = self.__problem.build_patch(patch)
            logger.debug("built candidate: %s", candidate)
            self.outcomes.record_build(candidate, True, timer() - time_build_start)

            # for now, execute all tests in no particular order
            # TODO perform test ordering
            logger_c = logger.getChild(container.uid)
            logger_c.debug("executing tests")
            for test in self.__problem.tests:
                test_line_coverage = line_coverage_by_test[test.name]
                if not any(line in test_line_coverage for line in lines_changed):
                    logger_c.debug("skipping test: %s (%s)",
                                   test.name, candidate)
                    continue

                logger_c.debug("executing test: %s (%s)", test.name, candidate)
                self.__counter_tests += 1
                outcome = bz.containers.test(container, test)
                logger_c.debug("* test outcome: %s (%s) [retcode=%d]\n$ %s\n%s",
                               test.name,
                               candidate,
                               outcome.response.code,
                               test.command,
                               outcome.response.output)
                self.outcomes.record_test(candidate, test.name, outcome)
                if not outcome.passed:
                    logger_c.debug("* test failed: %s (%s)", test.name, candidate)
                    return True
                logger_c.debug("* test passed: %s (%s)", test.name, candidate)

            # if we've found a repair, pause the search
            self.__found_patches.append(candidate)
            logger_c.info("FOUND A REPAIR: %s", candidate)
            return True

        except BuildFailure:
            logger.debug("failed to build candidate: %s", candidate)
            self.outcomes.record_build(candidate, False, timer() - time_build_start)
            return True
        finally:
            logger.info("evaluated candidate: %s", candidate)
            if container is not None:
                del bz.containers[container.uid]



"""
This module is responsible for describing concrete transformations to source
code files.
"""
from typing import List, Iterator, Dict, FrozenSet, Tuple
import re
import logging

import attr
import rooibos
from bugzoo.core.bug import Bug

from .problem import Problem
from .snippet import Snippet
from .core import Replacement, FileLine, FileLocationRange

logger = logging.getLogger(__name__)

REGEX_HOLE = re.compile('(?<=:\[)\w+(?=\])')


def is_single_token(snippet: str) -> bool:
    return ' ' not in snippet


class Transformation(object):
    """
    Represents a transformation to a source code file.
    """
    def to_replacement(self, problem: Problem) -> Replacement:
        raise NotImplementedError


class RooibosTransformationMeta(type):
    def __new__(metacls: type, name: str, bases, dikt):
        if name != 'RooibosTransformation':
            if not 'match' in dikt:
                raise SyntaxError('missing "match" property in {}'.format(name))
            if not 'rewrite' in dikt:
                raise SyntaxError('missing "rewrite" property in {}'.format(name))
        # TODO add ability to specify constraints

        # determine free holes
        # holes_match = REGEX_HOLE.findall(dikt['match'])  # type: List[str]
        # holes_rewrite = REGEX_HOLE.findall(dikt['rewrite'])  # type: List[str]

        # synthesise a "is_valid_match" function for this transformation

        # TODO compose a fast rewrite function

        return type.__new__(metacls, name, bases, dikt)


@attr.s(frozen=True)
class RooibosTransformation(Transformation, metaclass=RooibosTransformationMeta):  # noqa: pycodestyle
    location = attr.ib(type=FileLocationRange)
    arguments = attr.ib(type=FrozenSet[Tuple[str, str]],  # TODO replace with FrozenDict
                        converter=lambda args: frozenset(args.items()))  # type: ignore  # noqa: pycodestyle

    # FIXME need to use abstract properties
    @property
    def rewrite(self) -> str:
        raise NotImplementedError

    @property
    def match(self) -> str:
        raise NotImplementedError

    def to_replacement(self, problem: Problem) -> Replacement:
        args = dict(self.arguments)
        text = problem.rooibos.substitute(self.rewrite, args)
        return Replacement(self.location, text)

    # FIXME implement using constraints
    @classmethod
    def is_valid_match(cls, match: rooibos.Match) -> bool:
        return True

    # FIXME automagically generate
    @classmethod
    def match_to_transformations(cls,
                                 problem: Problem,
                                 location: FileLocationRange,
                                 environment: rooibos.Environment
                                 ) -> List[Transformation]:
        args = {}  # type: Dict[str, str]
        return [cls(location, args)]  # type: ignore


class InsertVoidFunctionCall(RooibosTransformation):
    match = ";\n"
    rewrite = ";\n:[1]();\n"

    @classmethod
    def is_valid_match(cls, match: rooibos.Match) -> bool:
        # TODO must be inside a function
        return True

    @classmethod
    def match_to_transformations(cls,
                                 problem: Problem,
                                 location: FileLocationRange,
                                 environment: rooibos.Environment
                                 ) -> List[Transformation]:
        # don't insert into small functions

        # don't insert macros?

        # don't insert after a return statement (or a break?)
        line_previous = FileLine(location.filename, location.start.line)
        line_previous_content = \
            problem.sources.read_line(line_previous)
        if ' return ' in line_previous_content:
            return []

        # TODO find all unique insertion points

        # TODO find appropriate void functions
        args = {'1': 'foo'}
        return [cls(location, args)]


class InsertConditionalReturn(RooibosTransformation):
    match = ";\n"
    rewrite = ";\nif(:[1]){return;}\n"

    @classmethod
    def is_valid_match(cls, match: rooibos.Match) -> bool:
        # TODO must be inside a void function
        return True

    @classmethod
    def match_to_transformations(cls,
                                 problem: Problem,
                                 location: FileLocationRange,
                                 environment: rooibos.Environment
                                 ) -> List[Transformation]:
        # TODO contains_return
        # don't insert after a return statement (or a break?)
        line_previous = FileLine(location.filename, location.start.line)
        line_previous_content = \
            problem.sources.read_line(line_previous)
        if ' return ' in line_previous_content:
            return []

        # TODO find all unique insertion points

        # TODO find appropriate if guards
        args = {'1': 'true'}
        return [cls(location, args)]


class InsertConditionalBreak(RooibosTransformation):
    match = ";\n"
    rewrite = ";\nif(:[1]){break;}\n"

    @classmethod
    def is_valid_match(cls, match: rooibos.Match) -> bool:
        # TODO must be inside a loop
        return True

    @classmethod
    def match_to_transformations(cls,
                                 problem: Problem,
                                 location: FileLocationRange,
                                 environment: rooibos.Environment
                                 ) -> List[Transformation]:
        # TODO contains_return
        # don't insert after a return statement (or a break?)
        line_previous = FileLine(location.filename, location.start.line)
        line_previous_content = \
            problem.sources.read_line(line_previous)
        if ' return ' in line_previous_content:
            return []

        # TODO find all unique insertion points

        # TODO find appropriate if guards
        args = {'1': 'true'}
        return [cls(location, args)]


class ApplyTransformation(RooibosTransformation):
    match = "= :[1];"
    rewrite = "= :[2](:[1]);"

    @classmethod
    def is_valid_match(self, match: rooibos.Match) -> bool:
        return ';' not in match.environment['1'].fragment

    @classmethod
    def match_to_transformations(cls,
                                 problem: Problem,
                                 location: FileLocationRange,
                                 environment: rooibos.Environment
                                 ) -> List[Transformation]:
        # TODO find applicable transformations
        args = {'1': environment['1'].fragment,
                '2': 'foo'}  # type: Dict[str, str]
        return [cls(location, args)]  # type: ignore


class SignedToUnsigned(RooibosTransformation):
    match = "int :[1] ="
    rewrite = "unsigned int :[1] ="
    constraints = [
        ("1", is_single_token)
    ]

    @classmethod
    def is_valid_match(self, match: rooibos.Match) -> bool:
        return is_single_token(match.environment['1'].fragment)

    # FIXME borko?
    @classmethod
    def match_to_transformations(cls,
                                 problem: Problem,
                                 location: FileLocationRange,
                                 environment: rooibos.Environment
                                 ) -> List[Transformation]:
        args = {'1': environment['1'].fragment}  # type: Dict[str, str]
        return [cls(location, args)]  # type: ignore


class AndToOr(RooibosTransformation):
    match = "&&"
    rewrite = "||"


class OrToAnd(RooibosTransformation):
    match = "||"
    rewrite = "&&"


class LEToGT(RooibosTransformation):
    match = "<="
    rewrite = ">"


class GTToLE(RooibosTransformation):
    match = ">"
    rewrite = "<="


class LTToGE(RooibosTransformation):
    match = "<"
    rewrite = ">="


class GEToLT(RooibosTransformation):
    match = ">="
    rewrite = "<"


class EQToNEQ(RooibosTransformation):
    match = "=="
    rewrite = "!="


class NEQToEQ(RooibosTransformation):
    match = "!="
    rewrite = "=="


class PlusToMinus(RooibosTransformation):
    match = "+"
    rewrite = "-"


class MinusToPlus(RooibosTransformation):
    match = "-"
    rewrite = "+"


class DivToMul(RooibosTransformation):
    match = "/"
    rewrite = "*"


class MulToDiv(RooibosTransformation):
    match = "*"
    rewrite = "/"


@attr.s(frozen=True)
class LocationRangeTransformation(Transformation):
    location = attr.ib(type=FileLocationRange,
                       validator=attr.validators.instance_of(FileLocationRange))  # noqa: pycodestyle


@attr.s(frozen=True)
class DeleteTransformation(LocationRangeTransformation):
    def __str__(self) -> str:
        return "DELETE[{}]".format(self.location)

    def to_replacement(self, problem: Problem) -> Replacement:
        return Replacement(self.location, "")


@attr.s(frozen=True)
class ReplaceTransformation(LocationRangeTransformation):
    """
    Replaces a numbered line in a given file with a provided snippet.
    """
    snippet = attr.ib(type=Snippet)

    def __str__(self) -> str:
        return "REPLACE[{}; {}]".format(self.location, self.snippet)

    def to_replacement(self, problem: Problem) -> Replacement:
        return Replacement(self.location, self.snippet.content)


@attr.s(frozen=True)
class AppendTransformation(LocationRangeTransformation):
    """
    Appends a given snippet to a specific line in a given file.
    """
    snippet = attr.ib(type=Snippet)

    def __str__(self) -> str:
        return "APPEND[{}; {}]".format(self.location, self.snippet)

    def to_replacement(self, problem: Problem) -> Replacement:
        old = problem.sources.read_chars(self.location)
        new = old + self.snippet.content
        return Replacement(self.location, new)



__all__ = ['Metric', 'Localization']

from typing import Dict, Callable, List, Iterator, FrozenSet, Sequence
import random
import bisect

from bugzoo.core.spectra import Spectra

from .problem import Problem
from .core import FileLine
from .exceptions import NoImplicatedLines

Metric = Callable[[int, int, int, int], float]


class Localization(object):
    @staticmethod
    def build(problem: Problem,
              metric: Metric
              ) -> 'Localization':
        spectra = Spectra.from_coverage(problem.coverage)
        scores = {}  # type: Dict[FileLine, float]
        for line in spectra:
            row = spectra[line]
            scores[line] = metric(row.ep, row.np, row.ef, row.nf)
        return Localization(scores)

    def __init__(self, scores: Dict[FileLine, float]) -> None:
        """
        Raises:
            NoImplicatedLines: if no lines are determined to be suspicious.
            ValueError: if a line is assigned a negative suspiciousness.
        """
        self.__line_to_score = scores.copy()
        self.__lines = []  # type: List[FileLine]
        self.__scores = []  # type: List[float]
        for line, score in scores.items():
            if score < 0.0:
                raise ValueError("suspiciousness values must be non-negative.")
            if score == 0.0:
                continue
            self.__lines.append(line)
            self.__scores.append(score)
        self.__files = \
            frozenset(line.filename for line in self.__lines)  # type: FrozenSet[str]  # noqa: pycodestyle

        if not self.__lines:
            raise NoImplicatedLines

        # compute cumulative distribution function
        sm = sum(self.__scores)
        pdf = [s / sm for s in self.__scores]
        self.__cdf = [0.0]  # type: List[float]
        cum = pdf[0]
        for p in pdf[1:]:
            self.__cdf.append(cum)
            cum += p

    def __iter__(self) -> Iterator[FileLine]:
        yield from self.__lines

    def __getitem__(self, line: FileLine) -> float:
        return self.__line_to_score.get(line, 0.0)

    def without(self, line: FileLine) -> 'Localization':
        scores = self.__line_to_score.copy()
        del scores[line]
        return Localization(scores)

    def restricted_to_lines(self, lines: Sequence[FileLine]) -> 'Localization':
        scores = {l: s for (l, s) in self.__line_to_score.items()
                  if l in lines}
        return Localization(scores)

    def sample(self) -> FileLine:
        mu = random.random()
        i = bisect.bisect(self.__scores, mu)
        return self.__lines[i]

    def __len__(self) -> int:
        """
        Returns a count of the number of suspicious lines.
        """
        return len(self.__lines)

    @property
    def files(self) -> List[str]:
        """
        Returns a list of files that contain suspicious lines.
        """
        return list(self.__files)



"""
This module provides a number of composable methods for generating code
transformations and candidate patches.
"""
from typing import Iterator, List, Iterable, Tuple, Optional, Type, Dict
import random
import logging
import yaml

from rooibos import Client as RooibosClient
from rooibos import Match

from .localization import Localization
from .exceptions import NoImplicatedLines
from .core import FileLocationRange, FileLine, Location
from .problem import Problem
from .snippet import Snippet, SnippetDatabase
from .candidate import Candidate
from .transformation import Transformation, \
                            RooibosTransformation, \
                            AppendTransformation, \
                            DeleteTransformation, \
                            ReplaceTransformation

logger = logging.getLogger(__name__)


class Context(object):
    pass
    # provide access to files


class CandidateGenerator(Iterable):
    """
    Candidate generators are used to generate (normally in a lazy fashion) a
    stream of candidate patches. For now, candidate generators implement
    one-way communications: they do not accept inputs.
    """
    def __iter__(self) -> Iterator[Candidate]:
        return self

    def __next__(self) -> Candidate:
        raise NotImplementedError


class TransformationGenerator(Iterable):
    """
    Transformation generators are used to provide a stream of source code
    transformations. As with candidate generators, transformations are usually
    generated in a lazy fashion.
    """
    def __iter__(self) -> Iterator[Transformation]:
        return self

    def __next__(self) -> Transformation:
        raise NotImplementedError


class RooibosGenerator(TransformationGenerator):
    def __init__(self,
                 problem: Problem,
                 localization: Localization,
                 schemas: List[Type[RooibosTransformation]]
                 ) -> None:
        client_rooibos = problem.rooibos
        self.__problem = problem
        size = 0
        tally_by_file = \
            {fn: 0 for fn in localization.files}  # type: Dict[str, int]
        tally_by_schema = \
            {s: 0 for s in schemas}  # type: Dict[Type[RooibosTransformation], int]  # noqa: pycodestyle
        self.__localization = localization
        self.__transformations = \
            {l: {s: [] for s in schemas} for l in localization}  # type: Dict[FileLine, Dict[Type[RooibosTransformation], List[Transformation]]]  # noqa: pycodestyle

        logger.debug("computing transformations")
        for fn in localization.files:
            file_contents = problem.sources.read_file(fn)
            for schema in schemas:
                logger.debug("finding matches of %s in %s", schema.__name__, fn)
                tpl_match = schema.match
                matches = client_rooibos.matches(file_contents, tpl_match)
                for m in matches:
                    line = FileLine(fn, m.location.start.line)
                    if line not in localization:
                        continue
                    if not schema.is_valid_match(m):
                        logger.debug("skipping invalid match: %s", m)
                        continue
                    transformations = \
                        list(self._match_to_transformations(fn, schema, m))
                    size += len(transformations)
                    tally_by_schema[schema] += len(transformations)
                    tally_by_file[fn] += len(transformations)
                    self.__transformations[line][schema] += transformations

        # trim redundant parts of transformation map
        for line in localization:
            for schema in schemas:
                if not self.__transformations[line][schema]:
                    del self.__transformations[line][schema]
            if not self.__transformations[line]:
                del self.__transformations[line]

        # refine the fault localization to only cover represented lines
        lines = list(self.__transformations.keys())
        self.__localization = self.__localization.restricted_to_lines(lines)
        logger.info("finished computing transformations: %d transformations across %d lines",  # noqa: pycodestyle
                    size, len(self.__transformations))
        logger.info("transformations: %s", self.__transformations)
        logger.info('tranformations by schema:\n%s',
                    '\n'.join(['  * {}: {}'.format(s.__name__, count)
                               for (s, count) in tally_by_schema.items()]))
        logger.info('tranformations by file:\n%s',
            '\n'.join(['  * {}: {}'.format(fn, count)
                       for (fn, count) in tally_by_file.items()]))

    def _match_to_transformations(self,
                                  filename: str,
                                  schema: Type[RooibosTransformation],
                                  match: Match
                                  ) -> List[Transformation]:
        location = FileLocationRange(filename,
                                     Location(match.location.start.line,
                                              match.location.start.col),
                                     Location(match.location.stop.line,
                                              match.location.stop.col))
        return schema.match_to_transformations(self.__problem,
                                               location,
                                               match.environment)

    def __next__(self) -> Transformation:
        line = self.__localization.sample()
        logger.debug("looking for transformation at %s", line)
        operator_to_transformations = self.__transformations[line]

        # choose an operator at random
        # if there are no operator choices, discard this line
        # if no lines remain, we're finished
        try:
            op = random.choice(list(operator_to_transformations.keys()))
            transformations = operator_to_transformations[op]
        except IndexError:
            logger.debug("no transformations left at %s", line)
            del self.__transformations[line]
            try:
                self.__localization = self.__localization.without(line)
            except NoImplicatedLines:
                logger.debug("no transformations left in search space")
                raise StopIteration
            return self.__next__()

        # choose a transformation at random
        # if there are no more transformation choices, discard this operator
        # choice
        try:
            return transformations.pop()
        except IndexError:
            logger.debug("exhausted all %s transformations at %s", op, line)
            del operator_to_transformations[op]
            return self.__next__()


def all_transformations_in_file(
        problem: Problem,
        transformation_cls: Type[RooibosTransformation],
        filename: str
    ) -> Iterator[Transformation]:
    client_rooibos = problem.rooibos
    file_contents = problem.sources.read_file(filename)
    tpl_match = transformation_cls.match
    tpl_rewrite = transformation_cls.rewrite
    matches = client_rooibos.matches(file_contents, tpl_match)
    for m in matches:
        args = {}  # type: Dict[str, str]  # FIXME
        location = FileLocationRange(filename,
                                     Location(m.location.start.line,
                                              m.location.start.col),
                                     Location(m.location.stop.line,
                                              m.location.stop.col))
        yield transformation_cls(location, args)  # type: ignore


class SingleEditPatches(CandidateGenerator):
    """
    Provides a stream of single-transformation candidate patches composed using
    a provided stream of transformations.
    """
    def __init__(self,
                 transformations: Iterator[Transformation]
                 ) -> None:
        self.__transformations = transformations

    def __next__(self) -> Candidate:
        try:
            transformation = next(self.__transformations)
        except StopIteration:
            raise StopIteration
        return Candidate([transformation])  # type: ignore


class TargetSnippetGenerator(Iterable):
    def __init__(self,
                 targets: Iterator[FileLocationRange],
                 snippets: SnippetDatabase
                 ) -> None:
        self.__targets = targets
        self.__snippets = snippets
        self.__current_target = None # type: Optional[FileLocationRange]
        self.__snippets_at_target = iter([]) # type: Iterator[Snippet]

    def __iter__(self) -> Iterator[Tuple[FileLocationRange, Snippet]]:
        return self

    def __next__(self) -> Tuple[FileLocationRange, Snippet]:
        # fetch the next snippet at the current line
        # if there are no snippets left at this line, move onto
        # the next line. if there are no lines left, stop iterating.
        try:
            snippet = next(self.__snippets_at_target)
            return (self.__current_target, snippet)
        except StopIteration:
            try:
                # TODO use snippet generator here
                # - return snippets at current file
                # - add use/def restrictions
                self.__current_target = next(self.__targets)
                self.__snippets_at_target = \
                    self.__snippets.in_file(self.__current_target.filename)
                return self.__next__()
            except StopIteration:
                raise StopIteration


class DeletionGenerator(TransformationGenerator):
    def __init__(self, targets: Iterable[FileLocationRange]) -> None:
        """
        Constructs a deletion generator.
        """
        self.__targets = reversed(list(targets))

    def __next__(self) -> Transformation:
        """
        Returns the next deletion transformation from this generator.
        """
        try:
            next_target = next(self.__targets)
        except StopIteration:
            raise StopIteration

        # TODO add static analysis
        # should we delete this line?
        # * don't delete declarations
        return DeleteTransformation(next_target)


class ReplacementGenerator(TransformationGenerator):
    """
    Uses a provided snippet database to generate all legal replacement
    transformations for a sequence of transformation targets.
    """
    def __init__(self,
                 targets: Iterator[FileLocationRange],
                 snippets: SnippetDatabase
                 ) -> None:
        self.__generator_target_snippet = \
            TargetSnippetGenerator(targets, snippets)

    def __next__(self) -> Transformation:
        try:
            target, snippet = next(self.__generator_target_snippet)
        except StopIteration:
            raise StopIteration

        # TODO additional static analysis goes here
        # don't replace line with an equivalent

        return ReplaceTransformation(target, snippet)


class AppendGenerator(TransformationGenerator):
    def __init__(self,
                 targets: Iterator[FileLocationRange],
                 snippets: SnippetDatabase
                 ) -> None:
        self.__generator_target_snippet = \
            TargetSnippetGenerator(targets, snippets)

    def __next__(self) -> Transformation:
        try:
            target, snippet = next(self.__generator_target_snippet)
        except StopIteration:
            raise StopIteration

        # TODO additional static analysis goes here
        # * don't append after a return
        # * don't append after a break statement

        return AppendTransformation(target, snippet)


class AllTransformationsAtLine(TransformationGenerator):
    """
    Provides a stream of all the possible transformations that can be made at
    a single line using a given snippet database.
    """
    def __init__(self,
                 problem: Problem,
                 line: FileLine,
                 snippets: SnippetDatabase,
                 *,
                 randomize: bool = True
                 ) -> None:
        # transform line to character range
        # TODO tidy this hack
        char_range = problem.sources.line_to_location_range(line)

        # TODO clean up iterator ugliness
        self.__sources = [
            DeletionGenerator(iter([char_range])),
            ReplacementGenerator(iter([char_range]), snippets),
            AppendGenerator(iter([char_range]), snippets)
        ] # type: List[TransformationGenerator]

        # TODO implement randomize

    def __next__(self) -> Transformation:
        # TODO random/deterministic
        # choose a source at random
        try:
            source = random.choice(self.__sources)
        except IndexError:
            raise StopIteration

        # attempt to fetch a transformation from that source
        # if the source is exhausted, discard it and try again
        try:
            return next(source)
        except StopIteration:
            self.__sources.remove(source)
            return self.__next__()


# TODO: map from transformation targets to line numbers
class SampleByLocalization(TransformationGenerator):
    def __init__(self,
                 problem: Problem,
                 localization: Localization,
                 snippets: SnippetDatabase,
                 *,
                 randomize: bool = True
                 ) -> None:
        self.__localization = localization
        self.__transformations_by_line = {
            line: AllTransformationsAtLine(problem, line, snippets, randomize=randomize)
            for line in localization
        }

    def __next__(self) -> Transformation:
        try:
            line = self.__localization.sample()
            print("Looking at line: {}".format(line))
            source = self.__transformations_by_line[line]
        except ValueError:
            raise StopIteration

        try:
            return next(source)
        except StopIteration:
            del self.__transformations_by_line[line]
            self.__localization = self.__localization.without(line)
            return self.__next__()



from typing import Dict, Optional, Iterator
import attr

from bugzoo.core import TestOutcome as BugZooTestOutcome
from bugzoo.compiler import CompilationOutcome as BugZooBuildOutcome

from .candidate import Candidate

__all__ = [
    'TestOutcome',
    'TestOutcomeSet',
    'BuildOutcome',
    'CandidateOutcome',
    'OutcomeManager'
]


@attr.s(frozen=True)
class TestOutcome(object):
    """
    Records the outcome of a test execution.
    """
    successful = attr.ib(type=bool)
    time_taken = attr.ib(type=float)


@attr.s(frozen=True)
class BuildOutcome(object):
    """
    Records the outcome of a build attempt.
    """
    successful = attr.ib(type=bool)
    time_taken = attr.ib(type=float)


class TestOutcomeSet(object):
    """
    Records the outcome of different test executions for a single patch.
    """
    def __init__(self,
                 outcomes: Optional[Dict[str, TestOutcome]] = None
                 ) -> None:
        if outcomes is None:
            outcomes = {}
        self.__outcomes = outcomes # type: Dict[str, TestOutcome]

    def __iter__(self) -> Iterator[str]:
        return self.__outcomes.keys().__iter__()

    def __getitem__(self, test: str) -> TestOutcome:
        return self.__outcomes[test]

    def with_outcome(self, test: str, outcome: TestOutcome) -> 'TestOutcomeSet':
        outcomes = self.__outcomes.copy()
        outcomes[test] = outcome
        return TestOutcomeSet(outcomes)


@attr.s(frozen=True)
class CandidateOutcome(object):
    """
    Records the outcome of a candidate patch evaluation.
    """
    build = attr.ib(type=BuildOutcome)
    tests = attr.ib(type=TestOutcomeSet)

    def with_test_outcome(self,
                          test: str,
                          successful: bool,
                          time_taken: float
                          ) -> 'CandidateOutcome':
        outcome = TestOutcome(successful, time_taken)
        test_outcomes = self.tests.with_outcome(test, outcome)
        return CandidateOutcome(self.build, test_outcomes)


class OutcomeManager(object):
    # FIXME hash candidate outcomes
    def __init__(self) -> None:
        self.__outcomes = {} # type: Dict[Candidate, CandidateOutcome]

    def __getitem__(self, candidate: Candidate) -> CandidateOutcome:
        return self.__outcomes[candidate]

    def __iter__(self) -> Iterator[Candidate]:
        """
        Returns an iterator over the candidate patches whose outcomes are
        stored by this manager.
        """
        return self.__outcomes.keys().__iter__()

    def record_build(self,
                     candidate: Candidate,
                     successful: bool,
                     time_taken: float
                     ) -> None:
        outcome_build = BuildOutcome(successful, time_taken)
        c = CandidateOutcome(outcome_build, TestOutcomeSet())
        self.__outcomes[candidate] = c

    def record_test(self,
                    candidate: Candidate,
                    test_id: str,
                    test_outcome: BugZooTestOutcome
                    ) -> None:
        # TODO  race condition if there can be simultaneous test evaluations
        #       for a given patch; for now, that's not possible.
        candidate_outcome = self.__outcomes[candidate]
        candidate_outcome = \
            candidate_outcome.with_test_outcome(test_id,
                                                test_outcome.passed,
                                                test_outcome.duration)
        self.__outcomes[candidate] = candidate_outcome



__all__ = ['Candidate']

from typing import List, Iterator, Dict, FrozenSet

import attr
from bugzoo.core.patch import Patch

from .core import Replacement, FileLine
from .problem import Problem
from .transformation import Transformation


@attr.s(frozen=True)
class Candidate(object):
    """
    Represents a candidate repair as a set of atomic program transformations.
    """
    transformations = attr.ib(type=FrozenSet[Transformation],
                              converter=frozenset)  # type: ignore  # bug in mypy (should be fixed in v0.610)  # noqa: pycodestyle

    def to_diff(self, problem: Problem) -> Patch:
        """
        Transforms this candidate patch into a concrete, unified diff.
        """
        replacements = \
            map(lambda t: t.to_replacement(problem), self.transformations)
        replacements_by_file = {}  # type: Dict[str, List[Replacement]]
        for rep in replacements:
            fn = rep.location.filename
            if fn not in replacements_by_file:
                replacements_by_file[fn] = []
            replacements_by_file[fn].append(rep)
        # FIXME order each collection of replacements by location
        return problem.sources.replacements_to_diff(replacements_by_file)

    def lines_changed(self, problem: Problem) -> List[FileLine]:
        """
        Returns a list of source lines that are changed by this candidate
        patch.
        """
        replacements = \
            map(lambda t: t.to_replacement(problem), self.transformations)
        locations = [rep.location for rep in replacements]
        lines = [FileLine(loc.filename, loc.start.line)
                 for loc in locations]
        return lines



from typing import List, Optional, Dict, Iterator, Callable, Set, Iterable
from timeit import default_timer as timer
import tempfile
import logging
import os

import boggart
import rooibos
import bugzoo
from rooibos import Client as RooibosClient
from bugzoo.core.fileline import FileLine, FileLineSet
from bugzoo.core.container import Container
from bugzoo.core.bug import Bug
from bugzoo.core.patch import Patch
from bugzoo.core.coverage import TestSuiteCoverage
from bugzoo.core.test import TestCase
from bugzoo.compiler import CompilationOutcome as BuildOutcome
from bugzoo.util import indent

from .source import ProgramSourceManager
from .util import get_file_contents
from .exceptions import NoFailingTests, NoImplicatedLines, BuildFailure


logger = logging.getLogger(__name__)  # type: logging.Logger


class Problem(object):
    """
    Used to provide a description of a problem (i.e., a bug), and to hold
    information pertinent to its solution (e.g., coverage, transformations).
    """
    def __init__(self,
                 bz: bugzoo.BugZoo,
                 bug: Bug,
                 coverage: TestSuiteCoverage,
                 *,
                 client_rooibos: Optional[RooibosClient] = None,
                 ) -> None:
        """
        Constructs a Darjeeling problem description.

        Params:
            bug: A description of the faulty program.
            coverage: Used to provide coverage information for the program
                under test.

        Raises:
            NoFailingTests: if the program under repair has no failing tests.
            NoImplicatedLines: if no lines are implicated by the coverage
                information and the provided suspiciousness metric.
        """
        self.__bug = bug
        self.__client_rooibos = client_rooibos
        self.__client_bugzoo = bz
        self.__coverage = coverage
        self._dump_coverage()

        # determine the passing and failing tests
        logger.debug("using test execution used to generate coverage to determine passing and failing tests")
        self.__tests_failing = set()  # type: Set[TestCase]
        self.__tests_passing = set()  # type: Set[TestCase]
        for test_name in self.__coverage:
            test = bug.harness[test_name]
            test_coverage = self.__coverage[test_name]
            if test_coverage.outcome.passed:
                self.__tests_passing.add(test)
            else:
                self.__tests_failing.add(test)

        logger.info("determined passing and failing tests")
        logger.info("* passing tests: %s",
                    ', '.join([t.name for t in self.__tests_passing]))
        logger.info("* failing tests: %s",
                    ', '.join([t.name for t in self.__tests_failing]))
        if not self.__tests_failing:
            raise NoFailingTests

        # FIXME huge bottleneck!
        # cache contents of the implicated files
        t_start = timer()
        logger.debug("storing contents of source code files")
        self.__sources = ProgramSourceManager(bz,
                                              client_rooibos,
                                              bug,
                                              files=self.implicated_files)
        logger.debug("stored contents of source code files (took %.1f seconds)",
                     timer() - t_start)

    def _dump_coverage(self) -> None:
        logger.debug("[COVERAGE]\n%s\n[/COVERAGE]",
                     indent(str(self.__coverage), 2))

    def __remove_redundant_sources(self) -> None:
        logger.debug("reducing memory footprint by discarding extraneous data")
        source_files = set(self.__sources.files)  # type: Set[str]
        covered_files = set(self.__coverage.lines.files)  # type: Set[str]
        extraneous_files = covered_files - source_files
        for fn in extraneous_files:
            del self.__sources[fn]
        logger.debug("finished reducing memory footprint")

    def build_patch(self,
                    patch: Patch,
                    builder: Optional[Callable[[Container], BuildOutcome]] = None
                    ) -> Container:
        """
        Provisions a container for a given patch and prepares it by building
        the source code.

        Parameters:
            patch: the patch for which a container should be built.
            builder: used to optionally provide a custom function for building
                the code inside the container. If no custom builder is
                provided, then the default one will be used instead.

        Returns:
            a ready-to-use container that contains a built version of the
            patched source code.

        Raises:
            BuildFailure: if the program failed to build.
        """
        mgr_ctr = self.__client_bugzoo.containers
        container = None
        try:
            container = mgr_ctr.provision(self.__bug)
            mgr_ctr.patch(container, patch)
            if builder is None:
                outcome = mgr_ctr.build(container)
            else:
                outcome = builder(container)
            # logger.debug("build outcome for %s:\n%s",
            #              candidate,
            #              outcome.response.output)
            # ensure the container is destroyed
            if not outcome.successful:
                raise BuildFailure
        except Exception:
            if container is not None:
                del mgr_ctr[container.uid]
            raise
        return container

    def restrict_to_files(self, filenames: List[str]) -> None:
        """
        Restricts the scope of the repair to the intersection of the files
        that are currently within the scope of the repair and a set of
        provided files.
        """
        logger.info("restricting repair to files: %s", filenames)
        self.__coverage = self.__coverage.restricted_to_files(filenames)
        logger.info("successfully restricted repair to given files.")
        self._dump_coverage()
        self.__remove_redundant_sources()
        self.validate()

    def restrict_to_lines(self, lines: Iterable[FileLine]) -> None:
        """
        Restricts the scope of the repair to the intersection of the current
        set of implicated lines and a provided set of lines.
        """
        self.__coverage = self.__coverage.restricted_to_lines(lines)
        self._dump_coverage()
        self.__remove_redundant_sources()
        self.validate()

    def restrict_with_filter(self,
                             fltr: Callable[[str], bool]
                             ) -> None:
        """
        Uses a given filter to remove certain lines from the scope of the
        repair.
        """
        f = lambda fl: fltr(self.__sources.read_line(fl))
        filtered = [l for l in self.lines if f(l)]  # type: List[FileLine]
        self.restrict_to_lines(filtered)

    def validate(self) -> None:
        """
        Ensures that this repair problem is valid. To be considered valid, a
        repair problem must have at least one failing test case and one
        implicated line.
        """
        lines = list(self.lines)
        files = list(self.implicated_files)
        logger.info("implicated lines [%d]:\n%s", len(lines), lines)
        logger.info("implicated files [%d]:\n* %s", len(files),
                    '\n* '.join(files))
        if len(lines) == 0:
            raise NoImplicatedLines

    @property
    def rooibos(self) -> RooibosClient:
        assert self.__client_rooibos is not None
        assert isinstance(self.__client_rooibos, rooibos.Client)
        return self.__client_rooibos

    @property
    def bug(self) -> Bug:
        """
        A description of the bug, provided by BugZoo.
        """
        return self.__bug

    @property
    def tests(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the tests for this problem.
        """
        yield from self.__tests_failing
        yield from self.__tests_passing

    @property
    def tests_failing(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the failing tests for this problem.
        """
        return self.__tests_failing.__iter__()

    @property
    def tests_passing(self) -> Iterator[TestCase]:
        """
        Returns an iterator over the passing tests for this problem.
        """
        return self.__tests_passing.__iter__()

    @property
    def coverage(self) -> TestSuiteCoverage:
        """
        Line coverage information for each test within the test suite for the
        program under repair.
        """
        return self.__coverage

    @property
    def lines(self) -> Iterator[FileLine]:
        """
        Returns an iterator over the lines that are implicated by the
        description of this problem.
        """
        return self.__coverage.failing.lines.__iter__()

    @property
    def implicated_files(self) -> Iterator[str]:
        return self.__coverage.lines.files

    @property
    def sources(self) -> ProgramSourceManager:
        """
        The source code files for the program under repair.
        """
        return self.__sources



class DarjeelingError(Exception):
    pass


class NoFailingTests(DarjeelingError):
    """
    The program under repair has no failing tests and therefore does not
    constitute a valid test-based repair problem.
    """


class NoImplicatedLines(DarjeelingError):
    """
    The program under repair has no lines that have been marked as suspicious
    by its fault localisation and coverage information.
    """


class BuildFailure(DarjeelingError):
    """
    The project failed to build.
    """



__all__ = ['ProgramSourceManager']

from typing import List, Union, Dict, Optional, Iterator, Iterable

import boggart
from rooibos import Client as RooibosClient
from bugzoo.client import Client as BugZooClient
from bugzoo.core.patch import Patch
from bugzoo.core.bug import Bug as Snapshot

from .core import Replacement, FileLine, FileLocationRange, Location


# FIXME add option to save to disk
class ProgramSourceManager(object):
    def __init__(self,
                 client_bugzoo: BugZooClient,
                 client_rooibos: Optional[RooibosClient],
                 snapshot: Snapshot,
                 files: Iterable[str]
                 ) -> None:
        # FIXME hacko
        self.__snapshot = snapshot
        SFM = boggart.server.sourcefile.SourceFileManager
        self.__mgr = SFM(client_bugzoo,
                         client_rooibos,
                         boggart.config.Operators())  # type: ignore  # noqa: pycodestyle
        self.__files = list(files)
        self.__mgr._fetch_files(snapshot, self.__files)
        for fn in files:
            self.__mgr.read_file(snapshot, fn)

    @property
    def files(self) -> Iterator[str]:
        yield from self.__files

    # FIXME move to boggart
    def line_to_location_range(self, line: FileLine) -> FileLocationRange:
        """
        Returns the range of characters that are covered by a given line.
        """
        content = self.__mgr.read_line(self.__snapshot, line)  # TODO include newline?
        return FileLocationRange(line.filename,
                                 Location(line.num, 0),
                                 Location(line.num, len(content)))

    def num_lines(self, fn: str) -> int:
        """
        Computes the number of lines in a given source file.
        """
        return self.__mgr.num_lines(self.__snapshot, fn)

    def read_file(self, fn: str) -> str:
        """
        Returns the contents of a given source file.
        """
        return self.__mgr.read_file(self.__snapshot, fn)

    def read_line(self, at: FileLine, *, keep_newline: bool = False) -> str:
        """
        Returns the contents of a given line.

        Parameters:
            at: the location of the line.
            keep_newline: if set to True, then any newline character at the
                end of the line will be kept. If set to False, the trailing
                newline character will be removed.
        """
        content = self.__mgr.read_line(self.__snapshot, at)
        return content.rstrip('\n') if keep_newline else content

    def read_chars(self, at: FileLocationRange) -> str:
        return self.__mgr.read_chars(self.__snapshot, at)

    def __delitem__(self, fn: str) -> None:
        """
        Removes information for a given file from this manager.
        """
        self.__files.remove(fn)
        self.__mgr._forget_file(self.__snapshot, fn)

    def replacements_to_diff(self,
                             file_to_replacements: Dict[str, List[Replacement]]
                             ) -> Patch:
        return self.__mgr.replacements_to_diff(self.__snapshot,
                                               file_to_replacements)
